"""
Command-line script for checking an installed SPKG in an installation tree ($SAGE_LOCAL, $SAGE_VENV).
"""

# ****************************************************************************
#       Copyright (C) 2017 Erik M. Bray <erik.m.bray@gmail.com>
#                     2022 Matthias Koeppe
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from __future__ import print_function

import glob
import json
import os
import shutil
import subprocess
import sys
import argparse
import warnings

from .env import SAGE_ROOT

pth = os.path
PKGS = pth.join(SAGE_ROOT, 'build', 'pkgs')
"""Directory where all spkg sources are found."""


def check_lib_auditwheel(f, verbose=False):
    from auditwheel.lddtree import lddtree
    for lib, info in lddtree(f)["libs"].items():
        if verbose:
            print('- {0}: {1}'.format(lib, info["realpath"]), file=sys.stderr)
        if info["realpath"] is None:
            raise RuntimeError('Shared library {0} needed by {1} is not found'
                               .format(lib, f))


def installcheck(spkg_name, sage_local, verbose=False):
    """
    Given a package name and path to an installation tree (SAGE_LOCAL or SAGE_VENV),
    check the installation of the package in that tree.
    """

    # The default path to this directory; however its value should be read
    # from the environment if possible
    spkg_inst = pth.join(sage_local, 'var', 'lib', 'sage', 'installed')

    # Find all stamp files for the package; there should be only one, but if
    # there is somehow more than one we'll work with the most recent one.
    pattern = pth.join(spkg_inst, '{0}-*'.format(spkg_name))
    stamp_files = sorted(glob.glob(pattern), key=pth.getmtime)

    if stamp_files:
        stamp_file = stamp_files[-1]
    else:
        stamp_file = None

    spkg_meta = {}
    if stamp_file:
        try:
            with open(stamp_file) as f:
                spkg_meta = json.load(f)
        except (OSError, ValueError):
            pass

    if 'files' not in spkg_meta:
        if stamp_file:
            print("Old-style or corrupt stamp file '{0}'"
                  .format(stamp_file), file=sys.stderr)
        else:
            print("Package '{0}' is currently not installed in '{1}'"
                  .format(spkg_name, sage_local), file=sys.stderr)
    else:
        files = spkg_meta['files']

        for f in files:
            f = os.path.join(sage_local, f)
            if f.endswith(('.so', '.dylib')):
                if verbose:
                    print("Checking shared library file '{0}'"
                          .format(f), file=sys.stderr)
                if sys.platform == 'darwin':
                    try:
                        from delocate.libsana import _tree_libs_from_libraries, _filter_system_libs
                    except ImportError:
                        warnings.warn('delocate is not available, so nothing is actually checked')
                    else:
                        _tree_libs_from_libraries([f],
                                                  lib_filt_func=_filter_system_libs,
                                                  copy_filt_func=lambda path: True)
                else:
                    try:
                        check_lib_auditwheel(f, verbose=False)
                    except ImportError:
                        warnings.warn('auditwheel is not available, so nothing is actually checked')
            elif f.endswith('-any.whl'):
                # pure Python wheel, nothing to check
                pass
            elif f.endswith('.whl'):
                if verbose:
                    print("Checking wheel file '{0}'"
                          .format(f), file=sys.stderr)
                if sys.platform == 'darwin':
                    try:
                        from delocate import wheel_libs
                    except ImportError:
                        warnings.warn('delocate is not available, so nothing is actually checked')
                    else:
                        wheel_libs(f)
                else:
                    try:
                        from delocate.tmpdirs import TemporaryDirectory
                        from delocate.tools import zip2dir
                    except ImportError:
                        warnings.warn('delocate is not available, so nothing is actually checked')
                    else:
                        try:
                            with TemporaryDirectory() as tmpdir:
                                zip2dir(f, tmpdir)
                                for dirpath, dirnames, basenames in os.walk(tmpdir):
                                    for base in basenames:
                                        if base.endswith('.so'):
                                            depending_path = os.path.realpath(os.path.join(dirpath, base))
                                            check_lib_auditwheel(depending_path, verbose=False)
                        except ImportError:
                            warnings.warn('auditwheel is not available, so nothing is actually checked')


def dir_type(path):
    """
    A custom argument 'type' for directory paths.
    """

    if path and not pth.isdir(path):
        raise argparse.ArgumentTypeError(
            "'{0}' is not a directory".format(path))

    return path


def spkg_type(pkg):
    """
    A custom argument 'type' for spkgs--checks whether the given package name
    is a known spkg.
    """
    pkgbase = pth.join(PKGS, pkg)

    if not pth.isdir(pkgbase):
        raise argparse.ArgumentTypeError(
            "'{0}' is not an spkg listed in '{1}'".format(pkg, PKGS))

    return pkg


def make_parser():
    """Returns the command-line argument parser for sage-spkg-installcheck."""

    doc_lines = __doc__.strip().splitlines()

    parser = argparse.ArgumentParser(
        description=doc_lines[0],
        epilog='\n'.join(doc_lines[1:]).strip(),
        formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('spkg', type=spkg_type, help='the spkg to check')
    parser.add_argument('sage_local', type=dir_type, nargs='?',
                        default=os.environ.get('SAGE_LOCAL'),
                        help='the path of the installation tree (default: the $SAGE_LOCAL '
                             'environment variable if set)')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='verbose output showing all files removed')
    parser.add_argument('--debug', action='store_true', help=argparse.SUPPRESS)

    return parser


def run(argv=None):
    parser = make_parser()

    args = parser.parse_args(argv if argv is not None else sys.argv[1:])

    if args.sage_local is None:
        print('Error: An installation tree must be specified either at the command '
              'line or in the $SAGE_LOCAL environment variable',
              file=sys.stderr)
        sys.exit(1)

    try:
        installcheck(args.spkg, args.sage_local,
                  verbose=args.verbose)
    except Exception as exc:
        print("Error during installcheck of '{0}': {1}".format(
            args.spkg, exc), file=sys.stderr)

        if args.debug:
            raise

        sys.exit(1)


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-
"""
Controller for the commandline actions

AUTHORS:

    - Volker Braun (2016): initial version
    - Thierry Monteil (2022): clean option to remove outdated source tarballs
"""


# ****************************************************************************
#       Copyright (C) 2016      Volker Braun <vbraun.name@gmail.com>
#                     2020-2024 Matthias Koeppe
#                     2022      Thierry Monteil
#                     2024      Marc Culler
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************


import os
import re
import logging
log = logging.getLogger()

from collections import defaultdict

from sage_bootstrap.package import Package
from sage_bootstrap.tarball import Tarball, FileNotMirroredError
from sage_bootstrap.updater import ChecksumUpdater, PackageUpdater
from sage_bootstrap.creator import PackageCreator
from sage_bootstrap.pypi import PyPiVersion, PyPiNotFound, PyPiError
from sage_bootstrap.fileserver import FileServer
from sage_bootstrap.expand_class import PackageClass
from sage_bootstrap.env import SAGE_DISTFILES


# Approximation of https://peps.python.org/pep-0508/#names dependency specification
dep_re = re.compile('^ *([-A-Z0-9._]+)', re.IGNORECASE)


class Application(object):

    def config(self):
        """
        Print the configuration

        $ sage --package config
        Configuration:
          * log = info
          * interactive = True
        """
        log.debug('Printing configuration')
        from sage_bootstrap.config import Configuration
        print(Configuration())

    def list_cls(self, *package_classes, **filters):
        """
        Print a list of all available packages

        $ sage --package list
        4ti2
        _bootstrap
        _develop
        [...]
        zlib

        $ sage -package list --has-file=spkg-configure.m4 :experimental:
        perl_term_readline_gnu

        $ sage -package list --has-file=spkg-configure.m4 --has-file=distros/debian.txt
        4ti2
        _develop
        _prereq
        [...]
        zlib
        """
        log.debug('Listing packages')
        pc = PackageClass(*package_classes, **filters)
        for pkg_name in pc.names:
            print(pkg_name)

    def properties(self, *package_classes, **kwds):
        """
        Show the properties of given packages

        $ sage --package properties --format shell maxima
        path_maxima='........./build/pkgs/maxima'
        version_with_patchlevel_maxima='5.46.0'
        type_maxima='standard'
        source_maxima='normal'
        trees_maxima='SAGE_LOCAL'
        """
        props = kwds.pop('props', ['path', 'version_with_patchlevel', 'type', 'source', 'trees', 'purl'])
        format = kwds.pop('format', 'plain')
        log.debug('Looking up properties')
        pc = PackageClass(*package_classes)
        for package_name in pc.names:
            package = Package(package_name)
            if len(pc.names) > 1:
                if format == 'plain':
                    print("{0}:".format(package_name))
            for p in props:
                value = getattr(package, p)
                if value is None:
                    if p.startswith('version'):
                        value = 'none'
                    else:
                        value = ''
                if format == 'plain':
                    print("        {0:28} {1}".format(p + ":", value))
                else:
                    print("{0}_{1}='{2}'".format(p, package_name, value))

    def dependencies(self, *package_classes, **kwds):
        """
        Find the dependencies given package names

        $ sage --package dependencies maxima --runtime --order-only --format=shell
        order_only_deps_maxima='info'
        runtime_deps_maxima='ecl'
        """
        types = kwds.pop('types', None)
        format = kwds.pop('format', 'plain')
        log.debug('Looking up dependencies')
        pc = PackageClass(*package_classes)
        if format in ['plain', 'rst']:
            if types is None:
                typesets = [['build', 'order_only', 'runtime']]
            else:
                typesets = [[t] for t in types]
        elif format == 'shell':
            if types is None:
                types = ['build', 'order_only', 'optional', 'runtime', 'check']
            typesets = [[t] for t in types]
        else:
            raise ValueError('format must be one of "plain", "rst", and "shell"')

        for package_name in pc.names:
            package = Package(package_name)
            if len(pc.names) > 1:
                if format == 'plain':
                    print("{0}:".format(package_name))
                    indent1 = "        "
                elif format == 'rst':
                    print("\n{0}\n{1}\n".format(package_name, "~" * len(package_name)))
                    indent1 = ""
            else:
                indent1 = ""

            for typeset in typesets:
                if len(typesets) > 1:
                    if format == 'plain':
                        print(indent1 + "{0}: ".format('/'.join(typeset)))
                        indent2 = indent1 + "        "
                    elif format == 'rst':
                        print("\n" + indent1 + ".. tab:: {0}\n".format('/'.join(typeset)))
                        indent2 = indent1 + "    "
                else:
                    indent2 = indent1

                deps = []
                for t in typeset:
                    deps.extend(getattr(package, 'dependencies_' + t))
                deps = sorted(set(deps))

                if format in ['plain', 'rst']:
                    for dep in deps:
                        if '/' in dep:
                            # Suppress dependencies on source files, e.g. of the form $(SAGE_ROOT)/..., $(SAGE_SRC)/...
                            continue
                        if dep == 'FORCE':
                            # Suppress FORCE
                            continue
                        if dep.startswith('$('):
                            # Dependencies like $(BLAS)
                            print(indent2 + "- {0}".format(dep))
                        elif format == 'rst' and Package(dep).has_file('SPKG.rst'):
                            # This RST label is set in src/doc/bootstrap
                            print(indent2 + "- :ref:`spkg_{0}`".format(dep))
                        else:
                            print(indent2 + "- {0}".format(dep))
                elif format == 'shell':
                    # We single-quote the values because dependencies
                    # may contain Makefile variable substitutions
                    print("{0}_deps_{1}='{2}'".format(t, package_name, ' '.join(deps)))

    def name(self, tarball_filename):
        """
        Find the package name given a tarball filename

        $ sage --package name pari-2.8-1564-gdeac36e.tar.gz
        pari
        """
        log.debug('Looking up package name for %s', tarball_filename)
        tarball = Tarball(os.path.basename(tarball_filename))
        print(tarball.package.name)

    def tarball(self, package_name):
        """
        Find the tarball filenames given a package name

        $ sage --package tarball pari
        pari-2.8-1564-gdeac36e.tar.gz
        """
        log.debug('Looking up tarball name for %s', package_name)
        package = Package(package_name)
        tarballs = package.tarballs()
        for key in tarballs:
            print(tarballs[key].filename)

    def apropos(self, incorrect_name):
        """
        Find up to 5 package names that are close to the given name

        $ sage --package apropos python
        Did you mean: cython, ipython, python2, python3, patch?
        """
        log.debug('Apropos for %s', incorrect_name)
        from sage_bootstrap.levenshtein import Levenshtein, DistanceExceeded
        levenshtein = Levenshtein(5)
        names = []
        for pkg in Package.all():
            try:
                names.append([levenshtein(pkg.name, incorrect_name), pkg.name])
            except DistanceExceeded:
                pass
        if names:
            names = sorted(names)[:5]
            print('Did you mean: {0}?'.format(', '.join(name[1] for name in names)))
        else:
            print('There is no package similar to {0}'.format(incorrect_name))
            print('You can find further packages at http://files.sagemath.org/spkg/')

    def commit(self, package_name, message=None):
        """
        Commit the changes to the Sage source tree for the given package
        """
        package = Package(package_name)
        if message is None:
            message = 'build/pkgs/{0}: Update to {1}'.format(package_name, package.version)
        os.system('git commit -m "{0}" {1}'.format(message, package.path))

    def update(self, package_name, new_version, url=None, commit=False):
        """
        Update a package. This modifies the Sage sources.

        $ sage --package update pari 2015 --url=http://localhost/pari/tarball.tgz
        """
        log.debug('Updating %s to %s', package_name, new_version)
        update = PackageUpdater(package_name, new_version)
        if url is not None or update.package.tarball_upstream_url_pattern:
            log.debug('Downloading %s', url)
            update.download_upstream(url)
        update.fix_checksum()
        if commit:
            self.commit(package_name)

    def update_latest(self, package_name, commit=False):
        """
        Update a package to the latest version. This modifies the Sage sources.
        """
        pkg = Package(package_name)
        if pkg.source not in ['normal', 'wheel']:
            log.debug('update_latest can only update normal and wheel packages; %s is a %s package' % (pkg, pkg.source))
            return
        dist_name = pkg.distribution_name
        if dist_name is None:
            log.debug('%s does not have Python distribution info in version_requirements.txt' % pkg)
            return
        if pkg.tarball_pattern.endswith('.whl'):
            source = 'wheel'
        else:
            source = 'pypi'
        try:
            pypi = PyPiVersion(dist_name, source=source)
        except PyPiNotFound:
            log.debug('%s is not a pypi package', dist_name)
            return
        else:
            pypi.update(pkg)
        if commit:
            self.commit(package_name)

    def update_latest_cls(self, package_name_or_class, commit=False):
        exclude = [
            'cypari'   # Name conflict
        ]
        # Restrict to normal Python packages
        pc = PackageClass(package_name_or_class, has_files=['checksums.ini', 'version_requirements.txt'])
        if not pc.names:
            log.warn('nothing to do (does not name a normal Python package)')
        for package_name in sorted(pc.names):
            if package_name in exclude:
                log.debug('skipping %s because of pypi name collision', package_name)
                continue
            try:
                self.update_latest(package_name, commit=commit)
            except PyPiError as e:
                log.warn('updating %s failed: %s', package_name, e)

    def download(self, package_name, allow_upstream=False, tags=()):
        """
        Download a package

        $ sage --package download pari
        Using cached file /home/vbraun/Code/sage.git/upstream/pari-2.8-2044-g89b0f1e.tar.gz
        /home/vbraun/Code/sage.git/upstream/pari-2.8-2044-g89b0f1e.tar.gz
        """
        log.debug('Downloading %s', package_name)
        package = Package(package_name)
        tarballs = package.tarballs()
        sorted_tarballs = []
        for tag in tags:
            for key in tarballs:
                if key is not None and tag in key:
                    sorted_tarballs.append(tarballs[key])
                    del tarballs[key]
                    break
        if not sorted_tarballs:
            sorted_tarballs = tarballs.values()
        for tarball in sorted_tarballs:
            tarball.download(allow_upstream=allow_upstream)
            print(tarball.upstream_fqn)

    def download_cls(self, *package_classes, **kwds):
        """
        Download a package or a class of packages
        """
        allow_upstream = kwds.pop('allow_upstream', False)
        on_error = kwds.pop('on_error', 'stop')
        tags = kwds.pop('tags', [])
        has_files = list(kwds.pop('has_files', []))
        pc = PackageClass(*package_classes, has_files=has_files + ['checksums.ini'], **kwds)

        def download_with_args(package):
            try:
                self.download(package, allow_upstream=allow_upstream, tags=tags)
            except FileNotMirroredError:
                if on_error == 'stop':
                    raise
                elif on_error == 'warn':
                    log.warn('Unable to download tarball of %s', package)
                else:
                    raise ValueError('on_error must be one of "stop" and "warn"')
        pc.apply(download_with_args)

    def upload(self, package_name):
        """
        Upload a package to the Sage mirror network

        $ sage --package upload pari
        Uploading /home/vbraun/Code/sage.git/upstream/pari-2.8-2044-g89b0f1e.tar.gz
        """
        package = Package(package_name)
        tarballs = package.tarballs()
        for key in tarballs:
            tarball = tarballs[key]
            if not os.path.exists(tarball.upstream_fqn):
                log.debug('Skipping %s because it does not exist locally', tarball)
                continue
            if not tarball.is_distributable():
                log.info('Skipping %s because it is marked as not distributable', tarball)
                continue
            log.info('Uploading %s', tarball.upstream_fqn)
            fs = FileServer()
            fs.upload(tarball)

    def upload_cls(self, package_name_or_class):
        pc = PackageClass(package_name_or_class)
        pc.apply(self.upload)
        fs = FileServer()
        log.info('Publishing')
        fs.publish()

    def fix_checksum_cls(self, *package_classes):
        """
        Fix the checksum of packages

        $ sage --package fix-checksum
        """
        pc = PackageClass(*package_classes, has_files=['checksums.ini'])
        pc.apply(self.fix_checksum)

    def fix_checksum(self, package_name):
        """
        Fix the checksum of a package

        $ sage --package fix-checksum pari
        Updating checksum of pari-2.8-2044-g89b0f1e.tar.gz
        """
        log.debug('Correcting the checksum of %s', package_name)
        update = ChecksumUpdater(package_name)
        pkg = update.package
        if not pkg.tarball_pattern:
            log.info('Ignoring {0} because it is not a normal package'.format(package_name))
            return
        if not os.path.exists(pkg.tarball.upstream_fqn):
            log.info('Ignoring {0} because tarball is not cached'.format(package_name))
            return
        if pkg.tarball.checksum_verifies(force_sha256=True):
            log.info('Checksum of {0} (tarball {1}) unchanged'.format(package_name, pkg.tarball_filename))
        else:
            log.info('Updating checksum of {0} (tarball {1})'.format(package_name, pkg.tarball_filename))
            update.fix_checksum()

    def create(self, package_name, version=None, tarball=None, pkg_type=None, upstream_url=None,
               description=None, license=None, upstream_contact=None, pypi=False, source=None,
               dependencies=None):
        """
        Create a package

        $ sage --package create foo --version 1.3 --tarball FoO-VERSION.tar.gz --type experimental

        $ sage --package create scikit_spatial --pypi --type optional

        $ sage --package create torch --pypi --source pip --type optional

        $ sage --package create jupyterlab_markup --pypi --source wheel --type optional
        """
        if package_name.startswith('pypi/'):
            package_name = 'pkg:' + package_name
        if package_name.startswith('pkg:pypi/'):
            pypi = True
            package_name = package_name[len('pkg:pypi/'):].lower().replace('-', '_').replace('.', '_')
        elif '-' in package_name:
            raise ValueError('package names must not contain dashes, use underscore instead')
        if pypi:
            if source is None:
                try:
                    if PyPiVersion(package_name, source='wheel', version=version).tarball.endswith('-none-any.whl'):
                        source = 'wheel'
                    else:
                        source = 'normal'
                except PyPiError:
                    source = 'normal'
            pypi_version = PyPiVersion(package_name, source=source, version=version)
            if source == 'normal':
                if not tarball:
                    # Guess the general format of the tarball name.
                    tarball = pypi_version.tarball.replace(pypi_version.version, 'VERSION')
                if not version:
                    version = pypi_version.version
                # Use a URL from pypi.io instead of the specific URL received from the PyPI query
                # because it follows a simple pattern.
                upstream_url = 'https://pypi.io/packages/source/{0:1.1}/{0}/{1}'.format(package_name, tarball)
            elif source == 'wheel':
                if not version:
                    version = pypi_version.version
                if dependencies is None:
                    log.info('Requires-Python: {0}'.format(pypi_version.requires_python or 'none'))
                    requires_dist = pypi_version.requires_dist
                    if requires_dist:
                        dependencies = []
                        for item in requires_dist:
                            if "extra ==" in item:
                                continue
                            try:
                                dep = dep_re.match(item).groups()[0].strip()
                            except Exception:
                                continue
                            dep = 'pkg:pypi/' + dep
                            try:
                                dep = Package(dep).name
                            except ValueError:
                                self.create(dep, pkg_type=pkg_type)
                                dep = Package(dep).name
                            dependencies.append(dep)
            if not description:
                description = pypi_version.summary
            if not license:
                license = pypi_version.license
            if not upstream_contact:
                upstream_contact = pypi_version.package_url
        if upstream_url and not tarball:
            tarball = upstream_url.rpartition('/')[2]
        if tarball and source is None:
            source = 'normal'
        if not pkg_type:
            pkg_type = 'optional'
        log.info('Creating build/pkgs/%s', package_name)
        creator = PackageCreator(package_name)
        if version:
            creator.set_version(version)
        if pkg_type:
            creator.set_type(pkg_type)
        if description or license or upstream_contact:
            creator.set_description(description, license, upstream_contact)
        if pypi or source == 'pip':
            creator.set_python_data_and_scripts(pypi_package_name=pypi_version.name, source=source,
                                                dependencies=dependencies)
        if source == 'wheel':
            creator.set_pypi_urls(pypi_version)
        if tarball:
            creator.set_tarball(tarball, upstream_url)
            if upstream_url and version:
                update = PackageUpdater(package_name, version)
                update.download_upstream()
            else:
                update = ChecksumUpdater(package_name)
            update.fix_checksum()

    def clean(self):
        """
        Remove outdated source tarballs from the upstream/ directory

        $ sage --package clean
        42 files were removed from the .../upstream directory
        """
        log.debug('Cleaning upstream/ directory')
        package_names = PackageClass(':all:').names
        keep = [Package(package_name).tarball.filename for package_name in package_names]
        count = 0
        for filename in os.listdir(SAGE_DISTFILES):
            if filename not in keep:
                filepath = os.path.join(SAGE_DISTFILES, filename)
                if os.path.isfile(filepath):
                    log.debug('Removing file {}'.format(filepath))
                    os.remove(filepath)
                    count += 1
        print('{} files were removed from the {} directory'.format(count, SAGE_DISTFILES))

    def metrics_cls(self, *package_classes):
        """
        Show the metrics of given packages

        $ sage --package metrics :standard:
        has_file_distros_arch_txt=131
        has_file_distros_conda_txt=216
        has_file_distros_debian_txt=125
        has_file_distros_fedora_txt=138
        has_file_distros_gentoo_txt=181
        has_file_distros_homebrew_txt=61
        has_file_distros_macports_txt=129
        has_file_distros_nix_txt=51
        has_file_distros_opensuse_txt=146
        has_file_distros_slackware_txt=25
        has_file_distros_void_txt=184
        has_file_patches=35
        has_file_spkg_check=59
        has_file_spkg_configure_m4=222
        has_file_spkg_install=198
        has_tarball_upstream_url=231
        line_count_file_patches=22561
        line_count_file_spkg_check=402
        line_count_file_spkg_configure_m4=2792
        line_count_file_spkg_install=2960
        packages=272
        type_standard=272
        """
        log.debug('Computing metrics')
        metrics = defaultdict(int)
        pc = PackageClass(*package_classes)
        for package_name in pc.names:
            package = Package(package_name)
            metrics['packages'] += 1
            metrics['type_' + package.type] += 1
            for filenames in [['spkg-configure.m4'],
                              ['spkg-install', 'spkg-install.in'],
                              ['spkg-check', 'spkg-check.in'],
                              ['distros/arch.txt'],
                              ['distros/conda.txt'],
                              ['distros/debian.txt'],
                              ['distros/fedora.txt'],
                              ['distros/gentoo.txt'],
                              ['distros/homebrew.txt'],
                              ['distros/macports.txt'],
                              ['distros/nix.txt'],
                              ['distros/opensuse.txt'],
                              ['distros/slackware.txt'],
                              ['distros/void.txt'],
                              ['patches']]:
                key = filenames[0].replace('.', '_').replace('-', '_').replace('/', '_')
                metrics['has_file_' + key] += int(any(package.has_file(filename)
                                                      for filename in filenames))
                if not key.startswith('distros_'):
                    metrics['line_count_file_' + key] += sum(package.line_count_file(filename)
                                                             for filename in filenames)
            metrics['has_tarball_upstream_url'] += int(bool(package.tarball_upstream_url_pattern))
        for key, value in sorted(metrics.items()):
            print('{0}={1}'.format(key, value))



# -*- coding: utf-8 -*-
"""
Test sage-download-file commandline utility
"""

# ****************************************************************************
#       Copyright (C) 2016 Volker Braun <vbraun.name@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import os
import unittest
import tempfile
import subprocess
import shutil
import logging

from sage_bootstrap.env import SAGE_DISTFILES
from sage_bootstrap.download.mirror_list import MIRRORLIST_FILENAME
from sage_bootstrap.package import Package
from test.capture import CapturedLog


log = logging.getLogger()


PATH = os.path.join(
    os.path.dirname(os.path.dirname(__file__)),
    'bin'
)


EXECUTABLE = os.path.join(PATH, 'sage-package')


class SagePackageTestCase(unittest.TestCase):

    def run_command(self, *args, **kwds):
        env = dict(os.environ)
        env.update(kwds.get('env', {}))
        env['PATH'] = PATH + os.pathsep + env['PATH']
        kwds.update(
            stdin=None,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=env,
        )
        log.debug('running {}: {}'.format(args, kwds))
        proc = subprocess.Popen(args, **kwds)
        stdout, stderr = proc.communicate()
        stdout = stdout.decode('utf-8')
        stderr = stderr.decode('utf-8')
        log.debug(u'stdout="{}", stderr="{}"'.format(stdout, stderr))
        rc = proc.returncode
        return (rc, stdout, stderr)

    def test_config(self):
        rc, stdout, stderr = self.run_command(EXECUTABLE, 'config')
        # Prints nothing to stderr
        self.assertEqual(stderr, '')
        # returns successfully
        self.assertEqual(rc, 0)
        # Prints to stdout
        self.assertTrue(stdout.startswith('Configuration:\n'))

    def test_list(self):
        rc, stdout, stderr = self.run_command(EXECUTABLE, 'list')
        # Prints nothing to stderr
        self.assertEqual(stderr, '')
        # returns successfully
        self.assertEqual(rc, 0)
        # Prints to stdout
        self.assertTrue('configure' in stdout.splitlines())

    def test_name(self):
        pkg = Package('configure')
        rc, stdout, stderr = self.run_command(EXECUTABLE, 'name', pkg.tarball_filename)
        # Prints nothing to stderr
        self.assertEqual(stderr, '')
        # returns successfully
        self.assertEqual(rc, 0)
        # Prints to stdout
        self.assertEqual(stdout.rstrip(), 'configure')

    def test_tarball(self):
        pkg = Package('configure')
        rc, stdout, stderr = self.run_command(EXECUTABLE, 'tarball', pkg.name)
        # Prints nothing to stderr
        self.assertEqual(stderr, '')
        # returns successfully
        self.assertEqual(rc, 0)
        # Prints to stdout
        self.assertEqual(stdout.rstrip(), pkg.tarball_filename)

    def test_apropos(self):
        rc, stdout, stderr = self.run_command(EXECUTABLE, 'apropos', 'python')
        # Prints nothing to stderr
        self.assertEqual(stderr, '')
        # returns successfully
        self.assertEqual(rc, 0)
        # Prints to stdout
        self.assertTrue(stdout.startswith('Did you mean:'))

    def test_download(self):
        pkg = Package('configure')
        with CapturedLog() as _:
            pkg.tarball.download()
        rc, stdout, stderr = self.run_command(EXECUTABLE, 'download', pkg.name)
        # Prints info to stderr
        self.assertTrue(stderr.startswith('Using cached file'))
        # returns successfully
        self.assertEqual(rc, 0)
        # Prints filename to stdout
        self.assertEqual(stdout.rstrip(), pkg.tarball.upstream_fqn)

    def test_update(self):
        pkg = Package('configure')
        # The confball never has a patchlevel since we are upstream...
        self.assertEqual(pkg.patchlevel, -1)
        rc, stdout, stderr = self.run_command(EXECUTABLE, 'update', pkg.name, pkg.version)
        # Prints nothing to stderr
        self.assertEqual(stderr, '')
        # returns successfully
        self.assertEqual(rc, 0)
        # Prints nothing to stdout
        self.assertEqual(stdout, '')

    def test_fix_checksum(self):
        pkg = Package('configure')
        rc, stdout, stderr = self.run_command(EXECUTABLE, 'fix-checksum', 'configure')
        # Prints nothing to stderr
        self.assertEqual(stderr, '')
        # returns successfully
        self.assertEqual(rc, 0)
        # Prints to stdout
        self.assertEqual(
            stdout.rstrip(),
            'Checksum of {0} (tarball {1}) unchanged'.format(pkg.name, pkg.tarball_filename))

    def test_create(self):
        tmp = tempfile.mkdtemp()
        with open(os.path.join(tmp, 'configure.ac'), 'w+') as f:
            f.write('test')
        os.mkdir(os.path.join(tmp, 'build'))
        os.mkdir(os.path.join(tmp, 'build', 'pkgs'))
        os.mkdir(os.path.join(tmp, 'upstream'))
        with open(os.path.join(tmp, 'upstream', 'Foo-13.5.tgz'), 'w+') as f:
            f.write('tarball content')
        rc, stdout, stderr = self.run_command(
            EXECUTABLE,
            'create', 'foo',
            '--version', '13.5',
            '--tarball', 'Foo-VERSION.tgz',
            '--type', 'standard',
            env=dict(SAGE_ROOT=tmp)
        )
        self.assertEqual(rc, 0)
        with open(os.path.join(tmp, 'build', 'pkgs', 'foo', 'package-version.txt')) as f:
            self.assertEqual(f.read(), '13.5\n')
        with open(os.path.join(tmp, 'build', 'pkgs', 'foo', 'type')) as f:
            self.assertEqual(f.read(), 'standard\n')
        with open(os.path.join(tmp, 'build', 'pkgs', 'foo', 'checksums.ini')) as f:
            self.assertEqual(
                f.read(),
                'tarball=Foo-VERSION.tgz\n' +
                'sha1=15d0e36e27c69bc758231f8e9add837f40a40cd0\n' +
                'md5=bc62fed5e35f31aeea2af95c00473d4d\n' +
                'cksum=1436769867\n'
            )
        shutil.rmtree(tmp)
